"""Module with data plugins that represent files of completed calculations jobs that have been stashed."""

# AUTO-GENERATED

# fmt: off

from .base import *
from .folder import *

__all__ = (
    'RemoteStashData',
    'RemoteStashFolderData',
)

# fmt: on
