###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with implementation of the storage backend using PostGreSQL and the disk-objectstore."""

# AUTO-GENERATED

# fmt: off

from .backend import *

__all__ = (
    'PsqlDosBackend',
)

# fmt: on
