"""Module with data plugins that represent remote resources and so effectively are symbolic links."""

# AUTO-GENERATED

# fmt: off

from .base import *
from .stash import *

__all__ = (
    'RemoteData',
    'RemoteStashData',
    'RemoteStashFolderData',
)

# fmt: on
