"""Module with custom implementations of :class:`click.Group`."""

# AUTO-GENERATED

# fmt: off

from .dynamic import *
from .verdi import *

__all__ = (
    'DynamicEntryPointCommandGroup',
    'VerdiCommandGroup',
)

# fmt: on
